<?php

namespace yifang\model\relation;

use yifang\model\Relation;
use Closure;
use yifang\utils\Str;
use yifang\db\BaseQuery as Query;
use yifang\model\Model;
use yifang\model\Pivot;
use yifang\db\Raw;
use Exception;

/**
 * 一对多关联类
 */
class HasMany extends Relation
{
    /**
     * 架构函数
     */
    public function __construct(Model $parent, $model, $foreignKey,  $localKey)
    {
        $this->parent     = $parent;
        $this->model      = $model;
        $this->foreignKey = $foreignKey;
        $this->localKey   = $localKey;
        $this->query      = (new $model)->db();

        if (get_class($parent) == $model) {
            $this->selfRelation = true;
        }
    }

    /**
     * 延迟获取关联数据
     * @return Collection
     */
    public function getRelation(array $subRelation = [], Closure $closure = null)
    {
        if ($closure) {
            $closure($this->getClosureType($closure));
        }
        if ($this->withLimit) {
            $this->query->limit($this->withLimit);
        }
        return $this->query
            ->where($this->foreignKey, $this->parent->{$this->localKey})
            ->relation($subRelation)
            ->select()
            ->setParent(clone $this->parent);
    }

    /**
     * 预载入关联查询
     * @return void
     */
    public function eagerlyResultSet(array &$resultSet,  $relation, array $subRelation, Closure $closure = null, array $cache = [])
    {
        $localKey = $this->localKey;
        $range    = [];
        foreach ($resultSet as $result) {
            // 获取关联外键列表
            if (isset($result->$localKey)) {
                $range[] = $result->$localKey;
            }
        }
        if (!empty($range)) {
            $data = $this->eagerlyOneToMany([
                [$this->foreignKey, 'in', $range],
            ], $subRelation, $closure, $cache);
            // 关联数据封装
            foreach ($resultSet as $result) {
                $pk = $result->$localKey;
                if (!isset($data[$pk])) {
                    $data[$pk] = [];
                }
                $result->setRelation($relation, $this->resultSetBuild($data[$pk], clone $this->parent));
            }
        }
    }

    /**
     * 预载入关联查询
     * @return void
     */
    public function eagerlyResult(Model $result,  $relation, array $subRelation = [], Closure $closure = null, array $cache = [])
    {
        $localKey = $this->localKey;
        if (isset($result->$localKey)) {
            $pk   = $result->$localKey;
            $data = $this->eagerlyOneToMany([
                [$this->foreignKey, '=', $pk],
            ], $subRelation, $closure, $cache);
            // 关联数据封装
            if (!isset($data[$pk])) {
                $data[$pk] = [];
            }
            $result->setRelation($relation, $this->resultSetBuild($data[$pk], clone $this->parent));
        }
    }

    /**
     * 关联统计
     * @return integer
     */
    public function relationCount(Model $result, Closure $closure = null,  $aggregate = 'count',  $field = '*',  &$name = null)
    {
        $localKey = $this->localKey;
        if (!isset($result->$localKey)) {
            return 0;
        }
        if ($closure) {
            $closure($this->getClosureType($closure), $name);
        }
        return $this->query->where($this->foreignKey, '=', $result->$localKey)->$aggregate($field);
    }

    /**
     * 创建关联统计子查询
     * @return string
     */
    public function getRelationCountQuery(Closure $closure = null,  $aggregate = 'count',  $field = '*',  &$name = null)
    {
        if ($closure) {
            $closure($this->getClosureType($closure), $name);
        }
        return $this->query->alias($aggregate . '_table')->whereExp($aggregate . '_table.' . $this->foreignKey, '=' . $this->parent->getTable() . '.' . $this->localKey)->fetchSql()->$aggregate($field);
    }

    /**
     * 一对多 关联模型预查询
     * @return array
     */
    protected function eagerlyOneToMany(array $where, array $subRelation = [], Closure $closure = null, array $cache = [])
    {
        $foreignKey = $this->foreignKey;
        $this->query->removeWhereField($this->foreignKey);
        // 预载入关联查询 支持嵌套预载入
        if ($closure) {
            $this->baseQuery = true;
            $closure($this->getClosureType($closure));
        }
        if ($this->withoutField) {
            $this->query->withoutField($this->withoutField);
        }
        $list = $this->query
            ->where($where)
            ->cache($cache[0] ?? false, $cache[1] ?? null, $cache[2] ?? null)
            ->with($subRelation)
            ->select();
        // 组装模型数据
        $data = [];
        foreach ($list as $set) {
            $key = $set->$foreignKey;
            if ($this->withLimit && isset($data[$key]) && count($data[$key]) >= $this->withLimit) {
                continue;
            }
            $data[$key][] = $set;
        }
        return $data;
    }

    /**
     * 保存（新增）当前关联数据对象
     * @return Model|false
     */
    public function save($data, $replace = true)
    {
        $model = $this->make();
        return $model->replace($replace)->save($data) ? $model : false;
    }

    /**
     * 创建关联对象实例
     * @param array|Model $data
     * @return Model
     */
    public function make($data = [])
    {
        if ($data instanceof Model) {
            $data = $data->getData();
        }
        // 保存关联表数据
        $data[$this->foreignKey] = $this->parent->{$this->localKey};
        return new $this->model($data);
    }

    /**
     * 批量保存当前关联数据对象
     * @return array|false
     */
    public function saveAll(iterable $dataSet, $replace = true)
    {
        $result = [];
        foreach ($dataSet as $key => $data) {
            $result[] = $this->save($data, $replace);
        }
        return empty($result) ? false : $result;
    }

    /**
     * 根据关联条件查询当前模型
     * @return Query
     */
    public function has( $operator = '>=',  $count = 1,  $id = '*',  $joinType = 'INNER', Query $query = null)
    {
        $table = $this->query->getTable();
        $model    = class_basename($this->parent);
        $relation = class_basename($this->model);
        if ('*' != $id) {
            $id = $relation . '.' . (new $this->model)->getPk();
        }
        $softDelete = $this->query->getOptions('soft_delete');
        $query      = $query ?: $this->parent->db()->alias($model);
        return $query->field($model . '.*')
            ->join([$table => $relation], $model . '.' . $this->localKey . '=' . $relation . '.' . $this->foreignKey, $joinType)
            ->when($softDelete, function ($query) use ($softDelete, $relation) {
                $query->where($relation . strstr($softDelete[0], '.'), '=' == $softDelete[1][0] ? $softDelete[1][1] : null);
            })
            ->group($relation . '.' . $this->foreignKey)
            ->having('count(' . $id . ')' . $operator . $count);
    }

    /**
     * 根据关联条件查询当前模型
     * @return Query
     */
    public function hasWhere($where = [], $fields = null,  $joinType = '', Query $query = null)
    {
        $table    = $this->query->getTable();
        $model    = class_basename($this->parent);
        $relation = class_basename($this->model);

        if (is_array($where)) {
            $this->getQueryWhere($where, $relation);
        } elseif ($where instanceof Query) {
            $where->via($relation);
        } elseif ($where instanceof Closure) {
            $where($this->query->via($relation));
            $where = $this->query;
        }

        $fields     = $this->getRelationQueryFields($fields, $model);
        $softDelete = $this->query->getOptions('soft_delete');
        $query      = $query ?: $this->parent->db();

        return $query->alias($model)
            ->group($model . '.' . $this->localKey)
            ->field($fields)
            ->join([$table => $relation], $model . '.' . $this->localKey . '=' . $relation . '.' . $this->foreignKey, $joinType)
            ->when($softDelete, function ($query) use ($softDelete, $relation) {
                $query->where($relation . strstr($softDelete[0], '.'), '=' == $softDelete[1][0] ? $softDelete[1][1] : null);
            })
            ->where($where);
    }

    /**
     * 执行基础查询（仅执行一次）
     * @access protected
     * @return void
     */
    protected function baseQuery()
    {
        if (empty($this->baseQuery)) {
            if (isset($this->parent->{$this->localKey})) {
                // 关联查询带入关联条件
                $this->query->where($this->foreignKey, '=', $this->parent->{$this->localKey});
            }

            $this->baseQuery = true;
        }
    }

}
