<?php

namespace yifang\response;

use yifang\container\Container;
use yifang\lib\Cookie;

/**
 * 响应类
 */
abstract class Response
{
    /**
     * 原始数据
     * @var mixed
     */
    public $data;   //插件中间件需要修改这个值,实现动态增减参数,protected=>public

    /**
     * 模板路径
     * @var mixed
     */
    protected $templatePath;

    /**
     * 当前contentType
     * @var string
     */
    protected $contentType = 'text/html';

    /**
     * 字符集
     * @var string
     */
    protected $charset = 'utf-8';

    /**
     * 状态码
     * @var integer
     */
    protected $code = 200;

    /**
     * 输出参数
     * @var array
     */
    protected $options = [];

    /**
     * header参数
     * @var array
     */
    protected $header = [];

    /**
     * 输出内容
     * @var string
     */
    protected $content = null;

    /**
     * Cookie对象
     * @var Cookie
     */
    protected $cookie;

    /**
     * Session对象
     */
    protected $session;

    /**
     * 初始化
     * @access protected
     * @param mixed $data 输出数据
     * @param int $code 状态码
     */
    protected function init($data = '', $code = 200)
    {
        $this->data($data);
        $this->code = $code;

        $this->contentType($this->contentType, $this->charset);
    }

    /**
     * 发送HTTP状态
     */
    public function code($code)
    {
        $this->code = $code;
        return $this;
    }

    /**
     * LastModified
     */
    public function lastModified($time)
    {
        $this->header['Last-Modified'] = $time;
        return $this;
    }

    /**
     * Expires
     */
    public function expires($time)
    {
        $this->header['Expires'] = $time;
        return $this;
    }

    /**
     * ETag
     */
    public function eTag($eTag)
    {
        $this->header['ETag'] = $eTag;
        return $this;
    }

    /**
     * 创建Response对象
     */
    public static function create($data = '', $type = 'html', $code = 200)
    {
        $class = false !== strpos($type, '\\') ? $type : '\\yifang\\response\\' . ucfirst(strtolower($type));
        return Container::getInstance()->invokeClass($class, [$data, $code]);
    }

    /**
     * 设置响应头
     */
    public function header(array $header = [])
    {
        $this->header = array_merge($this->header, $header);
        return $this;
    }

    /**
     * 输出的参数
     * @access public
     * @param mixed $options 输出参数
     * @return $this
     */
    public function options(array $options = [])
    {
        $this->options = array_merge($this->options, $options);

        return $this;
    }

    /**
     * 发送数据到客户端
     * @access public
     * @return void
     * @throws \InvalidArgumentException
     */
    public function send()
    {
        $data = $this->getContent();
        if (!headers_sent()) {
            if (!empty($this->header)) {
                // 发送状态码
                http_response_code($this->code);
                // 发送头部信息
                foreach ($this->header as $name => $val) {
                    header($name . (!is_null($val) ? ':' . $val : ''));
                }
            }
            if ($this->cookie) {
                $this->cookie->save();
            }
        }

        $this->sendData($data);

        if (function_exists('fastcgi_finish_request')) {
            // 提高页面响应
            fastcgi_finish_request();
        }
    }

    /**
     * 设置页面输出内容
     */
    public function content($content)
    {
        if (null !== $content && !is_string($content) && !is_numeric($content) && !is_callable([$content,'__toString',])) {
            throw new \InvalidArgumentException(sprintf('variable type error： %s', gettype($content)));
        }
        $this->content = (string) $content;
        return $this;
    }

    /**
     * 设置Cookie
     */
    public function cookie($name, $value, $option = null)
    {
        $this->cookie->set($name, $value, $option);
        return $this;
    }

    /**
     * 获取输出数据
     * @access public
     * @return string
     */
    public function getContent()
    {
        if (null == $this->content) {
            $content = $this->output($this->data, $this->templatePath);
            if (null !== $content && !is_string($content) && !is_numeric($content) && !is_callable([$content, '__toString',])) {
                throw new \InvalidArgumentException(sprintf('variable type error： %s', gettype($content)));
            }
            $this->content = (string)$content;
        }
        return $this->content;
    }

    /**
     * 获取头部信息
     */
    public function getHeader($name = '')
    {
        if (!empty($name)) {
            return $this->header[$name] ?? null;
        }
        return $this->header;
    }

    /**
     * 处理数据
     */
    protected function output($data, $templatePath)
    {
        return $data;
    }

    /**
     * 获取原始数据
     * @access public
     * @return mixed
     */
    public function getData()
    {
        return $this->data;
    }

    /**
     * 输出数据
     * @access protected
     * @param string $data 要处理的数据
     * @return void
     */
    protected function sendData($data)
    {
        echo $data;
    }

    /**
     * 页面缓存控制
     * @access public
     * @param string $cache 状态码
     * @return $this
     */
    public function cacheControl($cache)
    {
        $this->header['Cache-control'] = $cache;

        return $this;
    }

    /**
     * 页面输出类型
     */
    public function contentType($contentType, $charset = 'utf-8')
    {
        $this->header['Content-Type'] = $contentType . '; charset=' . $charset;
        return $this;
    }

    /**
     * 输出数据设置
     */
    public function data($data)
    {
        $this->data = $data;
        return $this;
    }

    /**
     * 输出数据设置模板路径
     * @access public
     * @param mixed $data 输出数据前缀
     * @return $this
     */
    public function templatePath($templatePath = "")
    {
        $default=config("view.skin","");
        $this->templatePath = $default?$templatePath.DIRECTORY_SEPARATOR.$default:$templatePath;
        $config=array(
            "view_path"=>$templatePath,
        );
        config($config,"view");
        return $this;
    }

    /**
     * 获取响应状态码
     * @access public
     * @return int
     */
    public function getCode()
    {
        return $this->code;
    }
}
